@@ -238,6 +238,109 @@ close_welcome_session (GdmWelcomeSession *welcome_session)
         return ret;
 }
 
+static void
+load_lang_config_file (const char  *config_file,
+                       const char **str_array)
+{
+        gchar         *contents = NULL;
+        gchar         *p;
+        gchar         *str_joinv;
+        gchar         *pattern;
+        gchar         *key;
+        gchar         *value;
+        gsize          length;
+        GError        *error;
+        GString       *line;
+        GRegex        *re;
+
+        g_return_if_fail (config_file != NULL);
+        g_return_if_fail (str_array != NULL);
+
+        if (!g_file_test (config_file, G_FILE_TEST_EXISTS)) {
+                g_debug ("Cannot access '%s'", config_file);
+                return;
+        }
+
+        error = NULL;
+        if (!g_file_get_contents (config_file, &contents, &length, &error)) {
+                g_debug ("Failed to parse '%s': %s",
+                         config_file,
+                         (error && error->message) ? error->message : "(null)");
+                g_error_free (error);
+                return;
+        }
+
+        if (!g_utf8_validate (contents, length, NULL)) {
+                g_warning ("Invalid UTF-8 in '%s'", config_file);
+                g_free (contents);
+                return;
+        }
+
+        str_joinv = g_strjoinv ("|", (char **) str_array);
+        if (str_joinv == NULL) {
+                g_warning ("Error in joined");
+                g_free (contents);
+                return;
+        }
+
+        pattern = g_strdup_printf ("(?P<key>(%s))=(\")?(?P<value>[^\"]*)?(\")?",
+                                   str_joinv);
+        error = NULL;
+        re = g_regex_new (pattern, 0, 0, &error);
+        g_free (pattern);
+        g_free (str_joinv);
+        if (re == NULL) {
+                g_warning ("Failed to regex: %s",
+                           (error && error->message) ? error->message : "(null)");
+                g_error_free (error);
+                g_free (contents);
+                return;
+        }
+
+        line = g_string_new ("");
+        for (p = contents; p && *p; p = g_utf8_find_next_char (p, NULL)) {
+                gunichar ch;
+                GMatchInfo *match_info = NULL;
+
+                ch = g_utf8_get_char (p);
+                if ((ch != '\n') && (ch != '\0')) {
+                        g_string_append_unichar (line, ch);
+                        continue;
+                }
+
+                if (line->str && g_utf8_get_char (line->str) == '#') {
+                        goto next_line;
+                }
+
+                if (!g_regex_match (re, line->str, 0, &match_info)) {
+                        goto next_line;
+                }
+
+                if (!g_match_info_matches (match_info)) {
+                        goto next_line;
+                }
+
+                key = g_match_info_fetch_named (match_info, "key");
+                value = g_match_info_fetch_named (match_info, "value");
+
+                if (key && *key && value && *value) {
+                        g_setenv (key, value, TRUE);
+                } else if (key && *key) {
+                        g_unsetenv (key);
+                }
+
+                g_free (key);
+                g_free (value);
+next_line:
+                g_match_info_free (match_info);
+                g_string_set_size (line, 0);
+        }
+
+        g_string_free (line, TRUE);
+        g_regex_unref (re);
+        g_free (contents);
+}
+
 static GPtrArray *
 get_welcome_environment (GdmWelcomeSession *welcome_session)
 {
@@ -253,6 +356,8 @@ get_welcome_environment (GdmWelcomeSession *welcome_session)
         };
         int i;
 
+        load_lang_config_file (LANG_CONFIG_FILE,
+                               (const char **) optional_environment);
         env = g_ptr_array_new ();
 
         /* create a hash table of current environment, then update keys has necessary */
@@ -319,7 +424,7 @@ get_welcome_environment (GdmWelcomeSession *welcome_session)
 
 
         g_hash_table_insert (hash, g_strdup ("PATH"), g_strdup (g_getenv ("PATH")));
-
+        g_hash_table_insert (hash, g_strdup ("WINDOWPATH"), g_strdup (g_getenv ("WINDOWPATH")));
         g_hash_table_insert (hash, g_strdup ("RUNNING_UNDER_GDM"), g_strdup ("true"));
         g_hash_table_insert (hash, g_strdup ("GVFS_DISABLE_FUSE"), g_strdup ("1"));
 
@@ -419,6 +524,7 @@ spawn_child_setup (SpawnChildData *data)
 {
         struct passwd *pwent;
         struct group  *grent;
+        int            res;
 
         if (data->user_name == NULL) {
                 return;
@@ -440,7 +546,11 @@ spawn_child_setup (SpawnChildData *data)
 
         g_debug ("GdmWelcomeSession: Setting up run time dir %s", data->runtime_dir);
         g_mkdir (data->runtime_dir, 0755);
-        chown (data->runtime_dir, pwent->pw_uid, pwent->pw_gid);
+        res = chown (data->runtime_dir, pwent->pw_uid, pwent->pw_gid);
+        if (res == -1) {
+                g_warning ("GdmWelcomeSession: Error setting owner of run time directory: %s",
+                           g_strerror (errno));
+        }
 
         g_debug ("GdmWelcomeSession: Changing (uid:gid) for child process to (%d:%d)",
                  pwent->pw_uid,
@@ -468,7 +578,7 @@ spawn_child_setup (SpawnChildData *data)
                 gid_t groups[1] = { 0 };
 
                 if (setgid (0) < 0)  {
-                        g_warning (_("Couldn't set groupid to 0"));
+                        g_warning (_("Couldn't set groupid to %d"), 0);
                         /* Don't error out, it's not fatal, if it fails we'll
                          * just still be */
                 }
@@ -503,6 +613,7 @@ static gboolean
 spawn_command_line_sync_as_user (const char *command_line,
                                  const char *user_name,
                                  const char *group_name,
+                                 const char *runtime_dir,
                                  const char *log_file,
                                  char       **env,
                                  char       **std_output,
@@ -528,6 +639,7 @@ spawn_command_line_sync_as_user (const char *command_line,
 
         data.user_name = user_name;
         data.group_name = group_name;
+        data.runtime_dir = runtime_dir;
         data.log_file = log_file;
 
         local_error = NULL;
@@ -662,7 +774,7 @@ parse_dbus_launch_output (const char *output,
         }
 
         if (addressp != NULL) {
-                *addressp = g_strdup (g_match_info_fetch (match_info, 1));
+                *addressp = g_match_info_fetch (match_info, 1);
         }
 
         if (pidp != NULL) {
@@ -699,10 +811,13 @@ start_dbus_daemon (GdmWelcomeSession *welcome_session)
 
         env = get_welcome_environment (welcome_session);
 
+        std_out = NULL;
+        std_err = NULL;
         error = NULL;
         res = spawn_command_line_sync_as_user (DBUS_LAUNCH_COMMAND,
                                                welcome_session->priv->user_name,
                                                welcome_session->priv->group_name,
+                                               welcome_session->priv->runtime_dir,
                                                NULL, /* log file */
                                                (char **)env->pdata,
                                                &std_out,
@@ -728,6 +843,8 @@ start_dbus_daemon (GdmWelcomeSession *welcome_session)
                 g_debug ("GdmWelcomeSession: Started D-Bus daemon on pid %d", welcome_session->priv->dbus_pid);
         }
  out:
+        g_free (std_out);
+        g_free (std_err);
         return res;
 }
 
@@ -855,7 +972,7 @@ gdm_welcome_session_stop (GdmWelcomeSession *welcome_session)
 
         g_debug ("GdmWelcomeSession: Stopping welcome_session");
 
-        res = gdm_signal_pid (-1 * welcome_session->priv->pid, SIGTERM);
+        res = gdm_signal_pid (welcome_session->priv->pid, SIGTERM);
         if (res < 0) {
                 g_warning ("Unable to kill welcome session process");
         } else {
@@ -1295,6 +1412,10 @@ gdm_welcome_session_finalize (GObject *object)
 
         gdm_welcome_session_stop (welcome_session);
 
+        if (welcome_session->priv->ckc != NULL) {
+                ck_connector_unref (welcome_session->priv->ckc);
+        }
+
         g_free (welcome_session->priv->command);
         g_free (welcome_session->priv->user_name);
         g_free (welcome_session->priv->group_name);
@@ -1306,6 +1427,8 @@ gdm_welcome_session_finalize (GObject *object)
         g_free (welcome_session->priv->server_address);
         g_free (welcome_session->priv->server_dbus_path);
         g_free (welcome_session->priv->server_dbus_interface);
+        g_free (welcome_session->priv->server_env_var_name);
+        g_free (welcome_session->priv->dbus_bus_address);
 
         G_OBJECT_CLASS (gdm_welcome_session_parent_class)->finalize (object);
 }
