@@ -37,6 +37,7 @@
 #include "gdm-greeter-client.h"
 #include "gdm-greeter-panel.h"
 #include "gdm-greeter-login-window.h"
+#include "gdm-user-chooser-widget.h"
 
 #include "gdm-profile.h"
 
@@ -101,12 +102,24 @@ on_reset (GdmGreeterClient  *client,
         gdm_greeter_login_window_reset (GDM_GREETER_LOGIN_WINDOW (session->priv->login_window));
 }
 
+static void
+show_or_hide_user_options (GdmGreeterSession *session,
+                           const char        *username)
+{
+    if (username != NULL && strcmp (username, GDM_USER_CHOOSER_USER_OTHER) != 0) {
+            gdm_greeter_panel_show_user_options (GDM_GREETER_PANEL (session->priv->panel));
+    } else {
+            gdm_greeter_panel_hide_user_options (GDM_GREETER_PANEL (session->priv->panel));
+    }
+}
+
 static void
 on_selected_user_changed (GdmGreeterClient  *client,
                           const char        *text,
                           GdmGreeterSession *session)
 {
         g_debug ("GdmGreeterSession: selected user changed: %s", text);
+        show_or_hide_user_options (session, text);
 }
 
 static void
@@ -125,8 +138,8 @@ on_default_layout_name_changed (GdmGreeterClient  *client,
                                 GdmGreeterSession *session)
 {
         g_debug ("GdmGreeterSession: default layout name changed: %s", text);
-        gdm_greeter_panel_set_default_layout_name (GDM_GREETER_PANEL (session->priv->panel),
-                                                   text);
+        gdm_greeter_panel_set_keyboard_layout (GDM_GREETER_PANEL (session->priv->panel),
+                                               text);
 }
 
 static void
@@ -240,7 +253,7 @@ on_select_user (GdmGreeterLoginWindow *login_window,
                 const char            *text,
                 GdmGreeterSession     *session)
 {
-        gdm_greeter_panel_show_user_options (GDM_GREETER_PANEL (session->priv->panel));
+        show_or_hide_user_options (session, text);
         gdm_greeter_client_call_select_user (session->priv->client,
                                              text);
 }
@@ -313,13 +326,15 @@ toggle_panel (GdmGreeterSession *session,
                 GdkScreen  *screen;
                 int         monitor;
                 int         x, y;
+                gboolean    is_local;
 
                 display = gdk_display_get_default ();
                 gdk_display_get_pointer (display, &screen, &x, &y, NULL);
 
                 monitor = get_tallest_monitor_at_point (screen, x, y);
 
-                session->priv->panel = gdm_greeter_panel_new (screen, monitor);
+                is_local = gdm_greeter_client_get_display_is_local (session->priv->client);
+                session->priv->panel = gdm_greeter_panel_new (screen, monitor, is_local);
 
                 g_signal_connect_swapped (session->priv->panel,
                                           "language-selected",
